(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: VertexBlend.cpp
//
// Desc: Example code showing how to do a skinning effect, using the vertex
//       blending feature of Direct3D. Normally, Direct3D transforms each
//       vertex through the world matrix. The vertex blending feature,
//       however, uses mulitple world matrices and a per-vertex blend factor
//       to transform each vertex.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit VertexBlendUnit;

interface

uses
  Windows, Messages, SysUtils, d3dres, Resource,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

// C++Builder compile hack
{$NOINCLUDE Resource}
{$HPPEMIT '#include "Resource.hpp"'}

type
  //-----------------------------------------------------------------------------
  // Name: struct BLENDVERTEX
  // Desc: Custom vertex which includes a blending factor
  //-----------------------------------------------------------------------------
  TBlendVertex = packed record
    v           : TD3DXVector3; // Referenced as v0 in the vertex shader
    blend       : Single;       // Referenced as v1.x in the vertex shader
    n           : TD3DXVector3; // Referenced as v3 in the vertex shader
    tu, tv      : Single;       // Referenced as v7 in the vertex shader
  end;

const
  TBlendVertex_FVF = D3DFVF_XYZB1 or D3DFVF_NORMAL or D3DFVF_TEX1;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;
    m_pObject:          CD3DMesh;       // Object to use for vertex blending
    m_dwNumVertices:    DWORD;
    m_dwNumFaces:       DWORD;
    m_pVB: IDirect3DVertexBuffer9;
    m_pIB: IDirect3DIndexBuffer9;

    m_matUpperArm:      TD3DXMatrix;    // Vertex blending matrices
    m_matLowerArm:      TD3DXMatrix;

    m_pVertexShader: IDirect3DVertexShader9; // Vertex shader
    m_bUseVertexShader: Boolean;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
  end;

implementation


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;
  m_strWindowTitle    :=  'VertexBlend: Surface Skinning Example';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont             := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pObject           := CD3DMesh.Create();
  m_pVB               := nil;
  m_pIB               := nil;
  m_pVertexShader     := nil;
  m_bUseVertexShader  := False;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  vAxis: TD3DXVector3;
  vZero, vOne: TD3DXVector4;
  vLight: TD3DXVector4;

  fDiffuse, fAmbient: array[0..3] of Single;

  matWorld0Transpose, matWorld1Transpose: TD3DXMatrix;
  matView, matProj, matViewProj, matViewProjTranspose: TD3DXMatrix;
begin
  // Set the vertex blending matrices for this frame
  vAxis:= D3DXVector3(2+Sin(m_fTime*3.1), 2+Sin(m_fTime*3.3), Sin(m_fTime*3.5));
  D3DXMatrixRotationAxis(m_matLowerArm, vAxis, Sin(3*m_fTime));
  D3DXMatrixIdentity(m_matUpperArm);

  // Set the vertex shader constants. Note: outside of the blend matrices,
  // most of these values don't change, so don't need to really be set every
  // frame. It's just done here for clarity
  if m_bUseVertexShader then
  begin
    // Some basic constants
    vZero:= D3DXVector4Zero;
    vOne:=  D3DXVector4(1, 1, 1, 1);

    // Lighting vector (normalized) and material colors. (Use red light
    // to show difference from non-vertex shader case.)
    vLight:= D3DXVector4(0.5, 1.0, -1.0, 0.0);
    D3DXVec4Normalize(vLight, vLight);
    // fDiffuse[] :=  { 1.00f, 1.00f, 0.00f, 0.00f };
    TD3DXVector4(fDiffuse) :=  D3DXVector4(1.00, 1.00, 0.00, 0.00);
    // fAmbient[] :=  { 0.25f, 0.25f, 0.25f, 0.25f };
    TD3DXVector4(fAmbient) :=  D3DXVector4(0.25, 0.25, 0.25, 0.25);

    // Vertex shader operations use transposed matrices
    m_pd3dDevice.GetTransform(D3DTS_VIEW,       matView);
    m_pd3dDevice.GetTransform(D3DTS_PROJECTION, matProj);
    D3DXMatrixMultiply(matViewProj, matView, matProj);
    D3DXMatrixTranspose(matWorld0Transpose, m_matUpperArm);
    D3DXMatrixTranspose(matWorld1Transpose, m_matLowerArm);
    D3DXMatrixTranspose(matViewProjTranspose, matViewProj);

    // Set the vertex shader constants
    m_pd3dDevice.SetVertexShaderConstantF(  0, @vZero,    1);
    m_pd3dDevice.SetVertexShaderConstantF(  1, @vOne,     1);
    m_pd3dDevice.SetVertexShaderConstantF(  4, @matWorld0Transpose,   4);
    m_pd3dDevice.SetVertexShaderConstantF(  8, @matWorld1Transpose,   4);
    m_pd3dDevice.SetVertexShaderConstantF( 12, @matViewProjTranspose, 4);
    m_pd3dDevice.SetVertexShaderConstantF( 20, @vLight,   1);
    m_pd3dDevice.SetVertexShaderConstantF( 21, @fDiffuse, 1);
    m_pd3dDevice.SetVertexShaderConstantF( 22, @fAmbient, 1);
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then 
  begin
    if m_bUseVertexShader then
    begin
      m_pd3dDevice.SetFVF(TBlendVertex_FVF);
      m_pd3dDevice.SetVertexShader(m_pVertexShader);
      m_pd3dDevice.SetStreamSource(0, m_pVB, 0, SizeOf(TBlendVertex));
      m_pd3dDevice.SetIndices(m_pIB);
      m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, m_dwNumVertices,
                                        0, m_dwNumFaces);
    end else
    begin
      // Enable vertex blending using API
      m_pd3dDevice.SetVertexShader(nil);
      m_pd3dDevice.SetTransform(D3DTS_WORLD,  m_matUpperArm);
      m_pd3dDevice.SetTransform(D3DTS_WORLD1, m_matLowerArm);
      m_pd3dDevice.SetRenderState(D3DRS_VERTEXBLEND, D3DVBF_1WEIGHTS);

      // Display the object
      m_pObject.Render(m_pd3dDevice);
    end;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    if m_bUseVertexShader
    then m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255), 'Using vertex shader')
    else m_pFont.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255), 'Using D3DRS_VERTEXBLEND');

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
type
  TBlendVertexArray = array [0..0] of TBlendVertex;
var
  pVB: IDirect3DVertexBuffer9;
  pVertices: ^TBlendVertexArray;
  dwNumVertices: DWord;

  fMinX, fMaxX: Single;
  i: DWord;
  a: Single;

  strVertexShaderPath: array [0..511] of Char;
  pCode: ID3DXBuffer;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Load an object to render
  Result:= m_pObject.Create_(m_pd3dDevice, PChar('mslogo.x'));
  if FAILED(Result) then Exit;

  if ( ((m_dwCreateFlags and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
        (m_dwCreateFlags and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0) and
      (m_d3dCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then
  begin
    // No VS available, so don't try to use it or allow user to
    // switch to it
    m_bUseVertexShader := False;
    EnableMenuItem(GetMenu(m_hWnd), IDM_USEVERTEXSHADER, MF_GRAYED);
  end
  else if (m_d3dCaps.MaxVertexBlendMatrices < 2) then
  begin
    // No blend matrices available, so don't try to use them or
    // allow user to switch to them
    m_bUseVertexShader :=  True;
    EnableMenuItem(GetMenu(m_hWnd), IDM_USEVERTEXSHADER, MF_GRAYED);
  end else
  begin
    // Both techniques available, so default to blend matrices and
    // allow the user to switch techniques
    m_bUseVertexShader :=  False;
    EnableMenuItem(GetMenu(m_hWnd), IDM_USEVERTEXSHADER, MF_ENABLED);
  end;

  // Set a custom FVF for the mesh
  m_pObject.SetFVF(m_pd3dDevice, TBlendVertex_FVF);

  // Add blending weights to the mesh
  {}
    // Gain acces to the mesh's vertices
    dwNumVertices := m_pObject.GetSysMemMesh.GetNumVertices;
    m_pObject.GetSysMemMesh.GetVertexBuffer(pVB);
    pVB.Lock(0, 0, Pointer(pVertices), 0);

    // Calculate the min/max z values for all the vertices
    fMinX:=   1e10;
    fMaxX:=  -1e10;

    for i:= 0 to dwNumVertices - 1 do
    begin
      if (pVertices[i].v.x < fMinX) then fMinX :=  pVertices[i].v.x;
      if (pVertices[i].v.x > fMaxX) then fMaxX :=  pVertices[i].v.x;
    end;

    for i:= 0 to dwNumVertices - 1 do
    begin
      // Set the blend factors for the vertices
      a := (pVertices[i].v.x - fMinX) / (fMaxX - fMinX);
      pVertices[i].blend := 1.0 - Sin(a*D3DX_PI*1.0);
    end;

    // Done with the mesh's vertex buffer data
    pVB.Unlock;
    pVB:= nil;
  {}

  if (m_dwCreateFlags and D3DCREATE_SOFTWARE_VERTEXPROCESSING <> 0) or
     (m_d3dCaps.VertexShaderVersion >= D3DVS_VERSION(1,0)) then
  begin
    // Find the vertex shader file
    Result := DXUtil_FindMediaFileCb(strVertexShaderPath, SizeOf(strVertexShaderPath), 'Blend.vsh');
    if FAILED(Result) then Exit;

    // Assemble the vertex shader from the file
    Result := D3DXAssembleShaderFromFile(strVertexShaderPath, nil, nil, 0, @pCode, nil);
    if FAILED(Result) then Exit;

    // Create the vertex shader
    Result := m_pd3dDevice.CreateVertexShader(PDWORD(pCode.GetBufferPointer), m_pVertexShader);
    pCode := nil;
    if FAILED(Result) then Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
var
  matProj: TD3DXMatrix;
  fAspect: Single;

  vEyePt:    TD3DXVector3;
  vLookatPt: TD3DXVector3;
  vUpVec:    TD3DXVector3;
  matView: TD3DXMatrix;

  light: TD3DLight9;
begin
  m_pFont.RestoreDeviceObjects;

  // Restore mesh's local memory objects
  m_pObject.RestoreDeviceObjects(m_pd3dDevice);

  // Get access to the mesh vertex and index buffers
  m_pObject.GetLocalMesh.GetVertexBuffer(m_pVB);
  m_pObject.GetLocalMesh.GetIndexBuffer(m_pIB);
  m_dwNumVertices :=  m_pObject.GetLocalMesh.GetNumVertices;
  m_dwNumFaces    :=  m_pObject.GetLocalMesh.GetNumFaces;

  // Set miscellaneous render states
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $00404040);

  // Set the projection matrix
  fAspect :=  m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 10000.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set the app view matrix for normal viewing
  vEyePt    :=  D3DXVector3(0.0,-5.0,-10.0);
  vLookatPt :=  D3DXVector3(0.0, 0.0,  0.0);
  vUpVec    :=  D3DXVector3(0.0, 1.0,  0.0);

  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  // Create a directional light. (Use yellow light to distinguish from
  // vertex shader case.)
  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, -0.5, -1.0, 1.0);
  light.Diffuse.r :=  1.0;
  light.Diffuse.g :=  1.0;
  light.Diffuse.b :=  0.0;
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pObject.InvalidateDeviceObjects;

  m_pVB:= nil;
  m_pIB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pObject.Destroy_;
  
  m_pVertexShader := nil;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pObject);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  if (dwBehavior and D3DCREATE_PUREDEVICE <> 0) then Exit; // GetTransform doesn't work on PUREDEVICE

  // Check that the device supports at least one of the two techniques
  // used in this sample: either a vertex shader, or at least two blend
  // matrices and a directional light.

  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexShaderVersion >= D3DVS_VERSION(1,0)) then
    begin
      Result:= S_OK;
      Exit;
    end;
  end else
  begin
    // Software vertex processing always supports vertex shaders
    Result:= S_OK;
    Exit;
  end;

  // Check that the device can blend vertices with at least two matrices
  // (Software can always do up to 4 blend matrices)
  if (pCaps.MaxVertexBlendMatrices < 2) then Exit;

  // If this is a TnL device, make sure it supports directional lights
  if (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0 ) or
     (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0) then
  begin
    if (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS = 0) then Exit;
  end;
  
  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM ): LRESULT;
const
  IsChecked: array[False..True] of Integer = (MF_UNCHECKED, MF_CHECKED);   
begin
  // Handle menu commands
  if (WM_COMMAND = uMsg) then
  begin
    case LOWORD(wParam) of
      IDM_USEVERTEXSHADER:
      begin
        m_bUseVertexShader:= not m_bUseVertexShader;
        CheckMenuItem(GetMenu(hWnd), IDM_USEVERTEXSHADER, IsChecked[m_bUseVertexShader]);
      end;
    end
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

